/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid;

import static de.ozgcloud.document.bescheid.BescheidEventListener.*;
import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections.MapUtils;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.user.OzgCloudUserProfile;
import de.ozgcloud.apilib.user.OzgCloudUserProfileService;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.document.BescheidDocumentCreatedEvent;
import de.ozgcloud.document.Document;
import de.ozgcloud.document.DocumentManagerConfiguration;
import de.ozgcloud.document.DocumentService;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import de.ozgcloud.nachrichten.postfach.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachRemoteService;
import de.ozgcloud.nachrichten.postfach.StringBasedIdentifier;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextAuthenticationTokenTestFactory;
import de.ozgcloud.vorgang.callcontext.WithMockCustomUser;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CreateCommandRequest;
import de.ozgcloud.vorgang.files.OzgCloudFile;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import de.ozgcloud.vorgang.status.StatusChangedEvent;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@SpringBootTest(classes = VorgangManagerServerApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.vorgang-manager.address=in-process:test",
		"grpc.client.nachrichten-manager.address=in-process:test",
		"grpc.client.command-manager.address=in-process:test",
		"grpc.client.pluto.address=in-process:test",
		"ozgcloud.feature.bescheid.enable-dummy-document-processor=true",
})
@DataITCase
@DirtiesContext
@WithMockCustomUser
class BescheidITCase {

	private static final String USER_ID = "user-id";

	@Autowired
	private CommandService commandService;
	@SpyBean
	private TestEventListener bescheiTestEventListener;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private PostfachRemoteService postfachRemoteService;
	@MockBean
	@Qualifier(DocumentManagerConfiguration.USER_PROFILE_SERVICE_NAME)
	private OzgCloudUserProfileService documentUserProfileService;

	@Captor
	private ArgumentCaptor<BescheidDocumentCreatedEvent> bescheidDocumentCreatedEventCaptor;

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		when(postfachRemoteService.getPostfachType()).thenReturn(ServiceKontoTestFactory.TYPE);
		when(documentUserProfileService.getById(any())).thenReturn(OzgCloudUserProfile.builder().id(OzgCloudUserId.from(USER_ID)).build());
	}

	@Nested
	class TestCreateBescheidDocument {

		@Test
		void shouldCreateBescheidDocument() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

			commandService.createCommand(buildCreatedCommandRequest(vorgangId, bescheidItemId));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
				assertThat(getBescheidDocument()).isNotNull();
			});
		}

		@Test
		void shouldCreateDocumentFile() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

			commandService.createCommand(buildCreatedCommandRequest(vorgangId, bescheidItemId));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
				assertThat(getBescheidDocumentFile()).isNotNull();
			});
		}

		private VorgangAttachedItem createBescheidAttachedItem(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name()))
					.build();
		}

		private CreateCommandRequest buildCreatedCommandRequest(String vorgangId, String bescheidItemId) {
			return CreateCommandRequest.builder()
					.vorgangId(vorgangId)
					.relationId(bescheidItemId)
					.order(BescheidEventListener.CREATE_BESCHEID_DOCUMENT_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODY_KEY, "2024-01-01",
							GENEHMIGT_BODY_KEY, true))
					.build();
		}

		private OzgCloudFile getBescheidDocumentFile() {
			var fileId = getBescheidDocument().getItem().get(Document.FIELD_DOCUMENT_FILE).toString();
			return mongoOperations.findById(fileId, OzgCloudFile.class);
		}

		private VorgangAttachedItem getBescheidDocument() {
			var documentId = bescheidDocumentCreatedEventCaptor.getValue().getCreatedResource();
			return mongoOperations.findById(documentId, VorgangAttachedItem.class);
		}

	}

	@Nested
	class TestSendBescheidManually {

		@Captor
		private ArgumentCaptor<BescheidSentEvent> bescheidSentEventCaptor;
		@Captor
		private ArgumentCaptor<StatusChangedEvent> statusChangedEventCaptor;

		private VorgangAttachedItem bescheidItem;
		private String vorgangId;

		@BeforeEach
		void mock() {
			vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			bescheidItem = mongoOperations.save(createBescheidAttachedItemWithDocument(vorgangId));
		}

		@Test
		void shouldSetBescheidStatusToSent() {
			commandService.createCommand(buildCreateCommandRequest());

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				assertThat(loadBescheid(bescheidItem.getId()).getVersion()).isEqualTo(2);
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadBescheid(bescheidItem.getId()).getItem()).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name());
			});
		}

		@Test
		void shouldSetVorgangStatusToBeschieden() {
			commandService.createCommand(buildCreateCommandRequest());

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				assertThat(loadBescheid(bescheidItem.getId()).getVersion()).isEqualTo(2);
				verify(bescheiTestEventListener).onBescheidStatusChangedEvent(statusChangedEventCaptor.capture());
				assertThat(loadVorgang(vorgangId).getStatus()).isEqualTo(Vorgang.Status.BESCHIEDEN);
			});
		}

		private VorgangAttachedItem createBescheidAttachedItemWithDocument(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(
							Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name(),
							Bescheid.FIELD_SEND_BY, Bescheid.SendBy.MANUAL.name(),
							Bescheid.FIELD_BESCHEID_DOCUMENT, "documentId"))
					.build();
		}

		private CreateCommandRequest buildCreateCommandRequest() {
			return CreateCommandRequest.builder()
					.vorgangId(vorgangId)
					.relationId(bescheidItem.getId())
					.relationVersion(bescheidItem.getVersion())
					.order(BescheidEventListener.SEND_BESCHEID_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODY_KEY, "2024-01-01",
							GENEHMIGT_BODY_KEY, true))
					.build();
		}

		private VorgangAttachedItem loadBescheid(String bescheidId) {
			return mongoOperations.findById(bescheidId, VorgangAttachedItem.class);
		}

		private Vorgang loadVorgang(String vorgangId) {
			return mongoOperations.findById(vorgangId, Vorgang.class);
		}
	}

	@Nested
	class TestSendBescheidPostfachMail {

		private static final String NACHRICHT_SUBJECT = "nachrichtSubject";
		private static final String NACHRICHT_TEXT = "nachrichtText";

		@Captor
		private ArgumentCaptor<BescheidSentEvent> bescheidSentEventCaptor;
		@Captor
		private ArgumentCaptor<StatusChangedEvent> statusChangedEventCaptor;
		@Captor
		private ArgumentCaptor<PostfachNachricht> postfachNachrichtCaptor;

		private String vorgangId;
		private VorgangAttachedItem bescheidItem;

		@BeforeEach
		void init() {
			SecurityContextHolder.getContext().setAuthentication(CallContextAuthenticationTokenTestFactory.create());
			vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var documentFileId = mongoOperations.save(createBescheidDocumentAttachedItem(vorgangId)).getId();
			bescheidItem = mongoOperations.save(createBescheidAttachedItemWithDocument(vorgangId, documentFileId));
		}

		@AfterAll
		static void cleanup() {
			SecurityContextHolder.clearContext();
		}

		@Test
		void shouldSuccessfullyCompleteCommand() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadCommand(bescheidSentEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
			});
		}

		@Test
		void shouldSetBescheidStatusToSent() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadBescheid(bescheidItem.getId()).getItem()).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name());
			});
		}

		@Test
		void shouldSetSentBy() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				var sentInfoMap = MapUtils.getMap(loadBescheid(bescheidItem.getId()).getItem(), Bescheid.FIELD_SENT_INFO);
				assertThat(MapUtils.getString(sentInfoMap, SentInfo.FIELD_SENT_BY)).isNotBlank();
			});
		}

		@Test
		void shouldSetSentAt() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				var sentAtStr = MapUtils.getString(
						MapUtils.getMap(loadBescheid(bescheidItem.getId()).getItem(), Bescheid.FIELD_SENT_INFO), SentInfo.FIELD_SENT_AT);
				assertThat(ZonedDateTime.parse(sentAtStr)).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
			});
		}

		@Test
		void shouldSetVorgangStatusToBeschieden() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidStatusChangedEvent(statusChangedEventCaptor.capture());
				assertThat(loadCommand(statusChangedEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
				assertThat(loadVorgang(vorgangId).getStatus()).isEqualTo(Vorgang.Status.BESCHIEDEN);
			});
		}

		@DirtiesContext
		@Test
		void shouldSendPostfachNachricht() {
			var expectedPostfachAddress = PostfachAddressTestFactory.createBuilder()
					.identifier(StringBasedIdentifier.builder()
							.postfachId(de.ozgcloud.vorgang.servicekonto.PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE)
							.build())
					.serviceKontoType(ServiceKontoTestFactory.TYPE).build();

			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(postfachRemoteService).sendMessage(postfachNachrichtCaptor.capture());
				assertThat(postfachNachrichtCaptor.getValue().getPostfachAddress()).usingRecursiveComparison().isEqualTo(expectedPostfachAddress);
				assertThat(postfachNachrichtCaptor.getValue().getSubject()).isEqualTo(NACHRICHT_SUBJECT);
				assertThat(postfachNachrichtCaptor.getValue().getMailBody()).isEqualTo(NACHRICHT_TEXT);
			});
		}

		@Test
		void shouldSetClientAttribute() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadCommand(bescheidSentEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
				var vorgang = loadVorgang(vorgangId);
				assertThat(vorgang.getClientAttributes()).containsKey(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);

			});
		}

		private VorgangAttachedItem createBescheidAttachedItemWithDocument(String vorgangId, String bescheidDocumentId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(
							Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name(),
							Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name(),
							Bescheid.FIELD_BEWILLIGT, true,
							Bescheid.FIELD_BESCHEID_DOCUMENT, bescheidDocumentId,
							Bescheid.FIELD_NACHRICHT_SUBJECT, NACHRICHT_SUBJECT,
							Bescheid.FIELD_NACHRICHT_TEXT, NACHRICHT_TEXT))
					.build();
		}

		private VorgangAttachedItem createBescheidDocumentAttachedItem(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(DocumentService.DOCUMENT_ITEM_NAME)
					.item(Map.of(
							DocumentService.FIELD_DOCUMENT_TYPE, DocumentService.DOCUMENT_TYPE,
							DocumentService.FIELD_DOCUMENT_FILE, "documentId",
							DocumentService.FIELD_NACHRICHT_TEXT, "nachrichtText"))
					.build();
		}

		private CreateCommandRequest buildCreateCommandRequest(String vorgangId, VorgangAttachedItem bescheidItem) {
			return CreateCommandRequest.builder()
					.vorgangId(vorgangId)
					.relationId(bescheidItem.getId())
					.relationVersion(bescheidItem.getVersion())
					.order(SEND_BESCHEID_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODY_KEY, "2024-01-01",
							GENEHMIGT_BODY_KEY, true))
					.build();
		}

		private Command loadCommand(String commandId) {
			return mongoOperations.findById(commandId, Command.class);
		}

		private VorgangAttachedItem loadBescheid(String bescheidId) {
			return mongoOperations.findById(bescheidId, VorgangAttachedItem.class);
		}

		private Vorgang loadVorgang(String vorgangId) {
			return mongoOperations.findById(vorgangId, Vorgang.class);
		}

	}
}
